module.exports = {
  meta: {
    type: 'problem',
    docs: {
      description:
        'Disallow returning new objects or arrays directly or indirectly in Redux selectors.',
      category: 'Best Practices',
      recommended: true
    },
    messages: {
      noNewObjectArray:
        'Avoid returning new objects or arrays directly in Redux selectors.',
      noVariableObjectArray:
        'Avoid returning variables referencing new objects or arrays in Redux selectors.',
      avoidCallExpression:
        'Avoid creating new reference via array and object methods. Opt for createSelector',
      noDefaultParamReference:
        'Avoid using default parameter values that create new references if they are returned.'
    }
  },
  create(context) {
    const config = context.options[0] || {};
    function getCurrentFunctionScope(scope) {
      if (scope.type === 'function') {
        return scope;
      }

      for (const innerScope of scope.childScopes) {
        if (innerScope.type === 'function') {
          return innerScope;
        }
      }
    }

    /**
     * List of methods known to create new references.
     */
    const referenceCreatingMethods = new Set(
      [
        'map',
        'filter',
        'reduce',
        'slice',
        'concat',
        'flat',
        'flatMap',
        'sort', // Sort mutates, but if chained, it creates a reference.
        'splice',
        'Object.assign',
        'Object.keys',
        'Object.values',
        'Object.entries'
      ].concat(config.referenceCreatingMethods || [])
    );

    /**
     * Checks if a node is a CallExpression invoking a reference-creating method.
     */
    function isReferenceCreatingMethodCall(node) {
      if (
        node.type === 'CallExpression' &&
        node.callee.type === 'MemberExpression'
      ) {
        const methodName = node.callee.property.name;
        return referenceCreatingMethods.has(methodName);
      }
      return false;
    }

    /**
     * Check if function parameters have invalid default values.
     */

    function trackDefaultParamUsage(node, paramName) {
      const references = getCurrentFunctionScope(
        context.getScope()
      ).references.filter(ref => ref.identifier.name === paramName);
      return references.some(ref => {
        const parent = ref.identifier.parent;
        return (
          parent.type === 'ReturnStatement' ||
          parent.type === 'ArrowFunctionExpression'
        );
      });
    }

    function checkDefaultParamValues(node) {
      if (node.params && node.params.length > 0) {
        for (const param of node.params) {
          if (
            param.type === 'AssignmentPattern' && // Detect default parameter values
            (isNewObjectOrArray(param.right) ||
              isReferenceCreatingMethodCall(param.right))
          ) {
            const paramName = param.left.name;
            if (trackDefaultParamUsage(node, paramName)) {
              context.report({
                node: param.right,
                messageId: 'noDefaultParamReference'
              });
            }
          }
        }
      }
    }
    /**
     * Checks if a node represents a new object or array being created.
     */
    function isNewObjectOrArray(node) {
      // console.log('--------\nisNewObjectOrArraynode', node);
      return (
        node.type === 'ObjectExpression' || // Matches {}
        node.type === 'ArrayExpression' // Matches []
      );
    }
    /**
     * Check if a returned value is invalid (new object/array or variable referencing it).
     */
    function checkReturnValue(node) {
      // console.log('--------------\ncheckreturnvalue', node);
      if (isNewObjectOrArray(node)) {
        // Report direct new object or array return.
        // console.log('--------\ngoing to be repored node directly', node);
        context.report({
          node,
          messageId: 'noNewObjectArray'
        });
      } else if (isReferenceCreatingMethodCall(node)) {
        // Report reference-creating methods directly.
        context.report({
          node,
          messageId: 'avoidCallExpression'
        });
      } else if (node.type === 'Identifier') {
        // Check if variable references a new object or array.
        // const variable = findVariable(context.getScope(), node);
        const name = node.name;
        const currentScope = getCurrentFunctionScope(context.getScope());
        // console.log('------------\ncurrentscope', currentScope);
        const variable = currentScope.set.get(name);
        // console.log('---------\nvariable', variable);
        if (variable && variable.defs.length > 0) {
          variable.defs.forEach(def => {
            // console.log('----------\nvariable def', def);
            const defNode = def.node;
            // if(defNode.type === 'Parameter' && defNode.)
            // console.log('----------\ndefnode', defNode);
            const initDefNode = defNode.init;
            // console.log('----------\ninitdefnode', initDefNode);
            if (initDefNode && isReferenceCreatingMethodCall(initDefNode)) {
              context.report({
                node,
                messageId: 'avoidCallExpression'
              });
            } else if (initDefNode && isNewObjectOrArray(initDefNode)) {
              // console.log('--------\ngoing to be repored node', node);
              context.report({
                node,
                messageId: 'noVariableObjectArray'
              });
            }
          });
        }
      }
    }

    /**
     * Recursively check expressions for invalid returns.
     */
    function checkExpression(node) {
      if (!node) return;
      // console.log('-----------------\nnodecheckexpre', node);
      if (
        node.type === 'LogicalExpression' ||
        node.type === 'ConditionalExpression'
      ) {
        // Check left and right operands of logical or conditional expressions.
        // console.log(
        //   '----------------\nnode left or consequent',
        //   node.left,
        //   node.consequent
        // );
        checkExpression(node.consequent || node.left);
        // console.log(
        //   '-------------\nnode right alternate',
        //   node.right,
        //   node.alternate
        // );
        checkExpression(node.alternate || node.right);
      } else {
        checkReturnValue(node);
      }
    }

    /**
     * Analyze the body of a function or arrow function for invalid returns.
     */
    function checkFunctionBody(node) {
      // Check function parameters for default initializers.
      checkDefaultParamValues(node);

      if (node.body.type === 'BlockStatement') {
        // For block-bodied functions, analyze return statements.
        const returnStatements = node.body.body.filter(
          stmt => stmt.type === 'ReturnStatement'
        );
        returnStatements.forEach(stmt => checkExpression(stmt.argument));
      } else {
        // For implicit returns in arrow functions.
        checkExpression(node.body);
      }
    }

    return {
      FunctionDeclaration(node) {
        const path = context.getFilename();
        const options = context.options[0];
        const isRuleNeedToRun =
          process.env.NODE_ENV === 'test' ||
          (options &&
            options.filenameRegex &&
            options.filenameRegex.test(path));
        if (isRuleNeedToRun) {
          checkFunctionBody(node);
        }
      },
      VariableDeclarator(node) {
        const path = context.getFilename();
        const options = context.options[0];
        const isFunction =
          node.init &&
          (node.init.type === 'ArrowFunctionExpression' ||
            node.init.type === 'FunctionExpression');
        const isRuleNeedToRun =
          process.env.NODE_ENV === 'test' ||
          (options &&
            options.filenameRegex &&
            options.filenameRegex.test(path));
        if (isRuleNeedToRun && isFunction) {
          // console.log('node init', node.init);
          checkFunctionBody(node.init);
        }
      }
      // ['FunctionDeclaration CallExpression'](node) {
      //   const path = context.getFilename();
      //   const options = context.options[0];
      //   if (
      //     options &&
      //     options.filenameRegex &&
      //     options.filenameRegex.test(path)
      //   ) {
      //   }
      // },
      // ['VariableDeclarator ArrowFunctionExpression CallExpression'](node) {
      //   const path = context.getFilename();
      //   const options = context.options[0];
      //   if (
      //     options &&
      //     options.filenameRegex &&
      //     options.filenameRegex.test(path)
      //   ) {
      //   }
      // }
      // ['VariableDeclarator Function CallExpression'](node) {
      //   const path = context.getFilename();
      //   const options = context.options[0];
      //   if (
      //     options &&
      //     options.filenameRegex &&
      //     options.filenameRegex.test(path)
      //   ) {
      //   }
      // }
    };
  }
};
