/* eslint-disable */
/**
 * @fileoverview To show error when followed bad practices inside mapStateToProps of react redux applications
 * @author Sivanesh
 */
'use strict';

//------------------------------------------------------------------------------
// Rule Definition
//------------------------------------------------------------------------------

function isInReturn(node) {
  let n = node;

  const returnBodyCondn = n =>
    n.type === 'ArrowFunctionExpression' &&
    n.body &&
    (n.body.type === 'ObjectExpression' ||
      n.body.type === 'ConditionalExpression' ||
      n.body.type === 'LogicalExpression');

  while (n && !(n.type === 'ReturnStatement' || returnBodyCondn(n))) {
    n = n.parent;
  }

  if (n && n.type === 'ReturnStatement') {
    return true;
  } else if (n && returnBodyCondn(n)) {
    // const mapStateToProps = (state) => ({})
    if (
      n.parent.type === 'VariableDeclarator' &&
      n.parent.id.name === 'mapStateToProps'
    ) {
      return true;
    } else if (
      n.parent.type === 'CallExpression' &&
      n.parent.callee.name === 'connect'
    ) {
      // connect((state) => ({})
      return true;
    } else if (
      n.parent.type === 'CallExpression' &&
      n.parent.callee.name === 'useSelector'
    ) {
      return true;
    }
  } else {
    return false;
  }
}

// makemapstatetoprops case
// return statement use panni new reference, but return statement use panna la na
// || case with new object

module.exports = {
  meta: {
    docs: {
      description:
        'To show error when followed bad practices inside mapStateToProps of react redux applications',
      category: 'Performance',
      recommended: true
    },
    fixable: null, // or "code" or "whitespace"
    schema: null
  },

  create: function (context) {
    const config = context.options[0] || {};
    const vulnerableQuery = ':matches(ArrayExpression, ObjectExpression)';
    const mapStateToPropsIdNameQuery = '[id.name="mapStateToProps"]';
    const makeMapStateToPropsIdNameQuery = '[id.name="makeMapStateToProps"]';
    const connectCallQuery = 'CallExpression[callee.name="connect"]';
    const useSelectoreQuery = 'CallExpression[callee.name="useSelector"]';
    //----------------------------------------------------------------------
    // Helpers
    //----------------------------------------------------------------------

    function foundInitialization(node) {
      // Switching the process as inside return and outside return.
      if (!isInReturn(node)) {
        return;
      }

      // For All return statement's Objects.
      if (node.parent.type === 'ReturnStatement') return;

      // connect(state => ({}))
      if (
        node.parent.type === 'ArrowFunctionExpression' &&
        node.parent.parent &&
        node.parent.parent.callee &&
        node.parent.parent.callee.name === 'connect'
      ) {
        return;
      }

      // const mapStateToProps = state =>  ({name: 1})
      if (
        node.type === 'ObjectExpression' &&
        node.parent.type === 'ArrowFunctionExpression' &&
        node.parent &&
        node.parent.parent &&
        node.parent.parent.id &&
        node.parent.parent.id.name === 'mapStateToProps'
      ) {
        return;
      }

      // If return objects function call has any array or object
      if (node.parent.type === 'CallExpression') {
        return;
      }

      // const { name = {} } = state || {}
      if (
        node.parent.type === 'LogicalExpression' &&
        node.parent.parent &&
        node.parent.parent.id &&
        node.parent.parent.id.type === 'ObjectPattern'
      ) {
        return;
      }

      // const { name = {} } = state ? state:  {}
      if (
        node.parent.type === 'ConditionalExpression' &&
        node.parent.parent &&
        node.parent.parent.id &&
        node.parent.parent.id.type === 'ObjectPattern'
      ) {
        return;
      }

      context.report({
        node,
        message:
          'Causes performance issue as it creates a new reference each time when redux calls this connect.'
      });
    }

    const arrRefs = [
      ...new Set(
        [
          'concat',
          'from',
          'of',
          'filter',
          'flat',
          'flatMap',
          'map',
          'slice',
          'reduce'
        ].concat(config.arrRefs || [])
      )
    ];

    const objRefs = [
      ...new Set(
        ['assign', 'create', 'keys', 'values', 'entries'].concat(
          config.objRefs || []
        )
      )
    ];

    function mstpFunction(node) {
      if (!isInReturn(node)) {
        return;
      }

      if (node.callee.property && arrRefs.includes(node.callee.property.name)) {
        context.report({
          node,
          message:
            'Creates new Array reference. Hence causes performance issues.'
        });
      }

      if (node.callee.object && node.callee.object.name === 'Object') {
        if (
          node.callee.property &&
          objRefs.includes(node.callee.property.name)
        ) {
          context.report({
            node,
            message:
              'Creates new Object reference. Hence causes performance issues.'
          });
        }
      }
    }

    return {
      // mapStateToProps check
      [`FunctionDeclaration${mapStateToPropsIdNameQuery} ${vulnerableQuery}`]:
        foundInitialization,
      [`VariableDeclarator${mapStateToPropsIdNameQuery} ArrowFunctionExpression ${vulnerableQuery}`]:
        foundInitialization,
      [`FunctionDeclaration${mapStateToPropsIdNameQuery} CallExpression`]:
        mstpFunction,
      [`VariableDeclarator${mapStateToPropsIdNameQuery} ArrowFunctionExpression CallExpression`]:
        mstpFunction,
      [`${connectCallQuery} ArrowFunctionExpression ${vulnerableQuery}`]:
        foundInitialization,
      [`${connectCallQuery} FunctionExpression ${vulnerableQuery}`]:
        foundInitialization,
      [`${connectCallQuery} ArrowFunctionExpression CallExpression`]:
        mstpFunction,
      [`${connectCallQuery} FunctionExpression CallExpression`]: mstpFunction,

      // makeMapStateToPropsCheck
      [`FunctionDeclaration${makeMapStateToPropsIdNameQuery} ${vulnerableQuery}`]:
        foundInitialization,
      [`VariableDeclarator${makeMapStateToPropsIdNameQuery} ArrowFunctionExpression ${vulnerableQuery}`]:
        foundInitialization,
      [`FunctionDeclaration${makeMapStateToPropsIdNameQuery} ArrowFunctionExpression CallExpression`]:
        mstpFunction,
      [`VariableDeclarator${makeMapStateToPropsIdNameQuery} ArrowFunctionExpression ArrowFunctionExpression CallExpression`]:
        mstpFunction,

      // useSelector check
      [`${useSelectoreQuery} ArrowFunctionExpression CallExpression`]:
        mstpFunction,
      [`${useSelectoreQuery} FunctionExpression CallExpression`]: mstpFunction,
      [`${useSelectoreQuery} ArrowFunctionExpression ${vulnerableQuery}`]:
        foundInitialization,
      [`${useSelectoreQuery} FunctionExpression ${vulnerableQuery}`]:
        foundInitialization
    };
  }
};
