const isReactReduxConnect = require('../isReactReduxConnect');
const utils = require('../utils');

const reportNoSelector = function (context, node, name) {
  context.report({
    message: `mapStateToProps property "${name}" should use a selector function.`,
    node,
  });
};

const reportWrongName = function (context, node, propName, functionName, matching) {
  context.report({
    message: `mapStateToProps "${propName}"'s selector "${functionName}" does not match "${matching}".`,
    node,
  });
};



const reportUnexpectedParam = function (context, node, propName, functionName, index) {
  context.report({
    message: `mapStateToProps "${propName}"'s selector "${functionName}" parameter #${index} is not expected.`,
    node,
  });
};

const reportInvalidParams = function (context, node, propName, functionName, params, index) {
  context.report({
    message: `mapStateToProps "${propName}"'s selector "${functionName}" parameter #${index} should be "${params[index].name}".`,
    node,
  });
};

const checkProperties = function (context, properties, matching, expectedParams) {
  properties.forEach((prop) => {
    if (prop.value.type !== 'CallExpression') {
      reportNoSelector(context, prop, prop.key.name);
      return;
    }
    if (matching && !prop.value.callee.name.match(new RegExp(matching))) {
      reportWrongName(context, prop, prop.key.name, prop.value.callee.name, matching);
    }
    if (expectedParams) {
      const actualParams = prop.value.arguments;
      const propName = prop.key.name;
      const functionName = prop.value.callee.name;
      actualParams.forEach((param, i) => {
        if (!expectedParams[i]) {
          reportUnexpectedParam(context, prop, propName, functionName, i);
          return;
        }
        if (param.name !== expectedParams[i].name) {
          reportInvalidParams(context, prop, propName, functionName, expectedParams, i);
        }
      });
    }
  });
};

const check = function (context, node, matching, validateParams) {
  const returnNode = utils.getReturnNode(node);
  if (utils.isObject(returnNode)) {
    checkProperties(context, returnNode.properties, matching, validateParams && node.params);
  }
};


function isUseSelector(node, hookNames) {
  return hookNames.includes(node.callee.name);
}

function reportUseSelectorWrongName(context, node, functionName, matching) {
  context.report({
    message: `${node.callee.name} selector "${functionName}" does not match "${matching}".`,
    node,
  });
}

function reportUseSelectorNoSelector(context, node) {
  context.report({
    message: `${node.callee.name} should use a named selector function.`,
    node,
  });
}

const create = function (context) {
  const config = context.options[0] || {};
  let hookNames = ['useSelector', 'useAppSelector'];

    // Ensure hookNames is an array
    if (config.hook) {
      hookNames = Array.isArray(config.hook) ? config.hook : [config.hook];
    }
  return {
    VariableDeclaration(node) {
      node.declarations.forEach((decl) => {
        if (decl.id && decl.id.name === 'mapStateToProps') {
          if (decl.init && (
            decl.init.type === 'ArrowFunctionExpression' ||
            decl.init.type === 'FunctionExpression'
          )) {
            check(context, decl.init, config.matching, !(config.validateParams === false));
          }
        }
      });
    },
    FunctionDeclaration(node) {
      if (node.id && node.id.name === 'mapStateToProps') {
        check(context, node.body, config.matching, !(config.validateParams === false));
      }
    },
    CallExpression(node) {
      if (isUseSelector(node, hookNames)) {
        const selector = node.arguments && node.arguments[0];
        if (selector && (
          selector.type === 'ArrowFunctionExpression' ||
          selector.type === 'FunctionExpression')
        ) {
          reportUseSelectorNoSelector(context, node);
        } else if (
          selector && selector.type === 'Identifier' &&
          config.matching &&
          !selector.name.match(new RegExp(config.matching))
        ) {
          reportUseSelectorWrongName(context, node, selector.name, config.matching);
        }
      };

      if (isReactReduxConnect(node)) {
        const mapStateToProps = node.arguments && node.arguments[0];
        if (mapStateToProps && (
          mapStateToProps.type === 'ArrowFunctionExpression' ||
          mapStateToProps.type === 'FunctionExpression')
        ) {
          check(context, mapStateToProps, config.matching, !(config.validateParams === false));
        }
      }
    },
  };
};

module.exports = {
  create,
  meta: {
    schema: {
      matching: {
        type: 'string'
      }
    }
  },
};